import os
import re
import sys

from jinjaform import log, terraform


args = sys.argv[1:]
cwd = os.getcwd()
env = os.environ.copy()

for name in ('JINJAFORM_PROJECT_ROOT', 'JINJAFORM_TERRAFORM_BIN'):
    if not env.get(name):
        log.bad('{} environment variable missing', name)
        sys.exit(1)

project_root = os.environ['JINJAFORM_PROJECT_ROOT']
jinjaform_dir = os.path.join(cwd, '.jinjaform')
jinjaform_root = os.path.join(project_root, '.jinjaform')
terraform_bin = os.environ['JINJAFORM_TERRAFORM_BIN']
terraform_dir = os.path.join(cwd, '.terraform')

aws_profiles = set()
tf_vars = {}
s3_backend = {}
sessions = {}


def read(_done=set()):
    """
    Reads configuration from *.tf and *.tfvars files.

    """

    default_vars = {}

    for name in os.listdir():

        if name in _done:
            continue

        if name.lower().endswith('.tf'):

            inside_aws_provider = False
            inside_s3_backend = False
            inside_terraform = False
            inside_variable = None

            for line in terraform.fmt(terraform_bin, name).splitlines():

                if inside_aws_provider:
                    if line == '}':
                        inside_aws_provider = False
                    else:
                        match = re.match(r'\s+profile\s+=\s+"(.+)"$', line)
                        if match:
                            aws_profile = match.group(1)
                            aws_profiles.add(aws_profile)
                elif inside_terraform:
                    if inside_s3_backend:
                        if line == '  }':
                            inside_s3_backend = False
                        else:
                            match = re.match(r'\s+(.+?)\s+=\s+"?(.+?)"?$', line)
                            if match:
                                name = match.group(1)
                                value = match.group(2)
                                s3_backend[name] = value
                    else:
                        if line == '}':
                            inside_terraform = False
                        elif line == '  backend "s3" {':
                            inside_s3_backend = True
                elif inside_variable:
                    if line == '}':
                        inside_variable = None
                    else:
                        match = re.match(r'^\s+default\s+=\s+"?(.*?)"?$', line)
                        if match:
                            default_vars[inside_variable] = match.group(1)
                elif line == 'provider "aws" {':
                    inside_aws_provider = True
                elif line == 'terraform {':
                    inside_terraform = True
                else:
                    match = re.match(r'^variable "(.+)" {$', line)
                    if match:
                        inside_variable = match.group(1)

            _done.add(name)

        elif name.lower().endswith('.tfvars'):

            for line in terraform.fmt(terraform_bin, name).splitlines():
                match = re.match(r'^([^\s]+)\s+=\s+"?(.*?)"?$', line)
                if match:
                    name = match.group(1)
                    value = match.group(2)
                    tf_vars[name] = value

            _done.add(name)

    # Use default variable values if not specified in tfvars.
    for name in default_vars:
        if name not in tf_vars:
            tf_vars[name] = default_vars[name]



import os
import shutil

from jinjaform.config import env, jinjaform_root, terraform_dir


def setup():
    """
    Sets up the terraform root link, module cache and plugin cache.

    """

    os.makedirs(terraform_dir, exist_ok=True)

    # Create a .terraform/root symlink so that Terraform
    # code can access it using a relative path.
    root_link = os.path.join(terraform_dir, 'root')
    if os.path.exists(root_link):
        os.remove(root_link)
    os.symlink(os.environ['TERRAFORM_ROOT'], root_link)

    # Create a shared modules directory in the root directory.
    module_link = os.path.join(terraform_dir, 'modules')
    module_cache_dir = os.path.join(jinjaform_root, 'modules')
    os.makedirs(module_cache_dir, exist_ok=True)
    if os.path.islink(module_link):
        os.remove(module_link)
    elif os.path.exists(module_link):
        shutil.rmtree(module_link)
    os.symlink(module_cache_dir, module_link)

    # Create a shared plugin cache directory in the root directory.
    plugin_cache_dir = os.path.join(jinjaform_root, 'plugins')
    os.makedirs(plugin_cache_dir, exist_ok=True)
    env['TF_PLUGIN_CACHE_DIR'] = plugin_cache_dir



import os
import sys
import subprocess

from jinjaform import log
from jinjaform.config import args


def abort():
    if os.environ.get('GIT_FORCE') == '1':
        pass
    else:
        log.bad('git: refusing to continue')
        log.ok('hint: set GIT_FORCE=1 to bypass git checks')
        sys.exit(1)


def check():

    if 'apply' not in args:
        return

    branch = subprocess.check_output(['git', 'rev-parse', '--abbrev-ref', 'HEAD']).rstrip().decode('utf-8')
    if branch != 'master':
        log.bad('git: working in branch {}', branch)
        abort()
        return

    diff = subprocess.check_output(['git', 'status', '--porcelain']).rstrip().decode('utf-8')
    if diff:
        log.bad('git: working directory not clean')
        for line in diff.splitlines():
            log.bad('git: {}', line)
        abort()
        return

    log.ok('git: checking origin')
    subprocess.check_output(['git', 'remote', 'update'])
    local = subprocess.check_output(['git', 'rev-parse', 'master']).rstrip().decode('utf-8')
    remote = subprocess.check_output(['git', 'rev-parse', 'origin/master']).rstrip().decode('utf-8')
    if local != remote:
        status = subprocess.check_output(['git', 'status']).rstrip().decode('utf-8')
        for line in status.splitlines():
            if 'branch' in line and ('ahead' in line or 'behind' in line):
                log.bad('git: {}', line.lower())
                break
        else:
            log.bad('git: master is out of date')
        abort()
        return



import os
import shutil
import sys

from jinjaform import log


shim_template = '''
#!/usr/bin/env sh
export JINJAFORM_PROJECT_ROOT={project_root}
export JINJAFORM_TERRAFORM_BIN={terraform_bin}
python -m jinjaform $*
'''


def install():

    if len(sys.argv) > 2:
        log.bad('usage: python -m jinjaform.direnv [project_root]')
        sys.exit(1)

    if len(sys.argv) == 2:
        project_root = os.path.abspath(sys.argv[1])
    else:
        project_root = os.getcwd()

    terraform_bin = shutil.which('terraform')
    if not terraform_bin:
        log.bad('usage: could not find terraform in path')
        sys.exit(1)

    jinjaform_root = os.path.join(project_root, '.jinjaform')

    for name in ('bin', 'modules', 'plugins'):
        os.makedirs(os.path.join(jinjaform_root, name), exist_ok=True)

    shim_path = os.path.join(jinjaform_root, 'bin', 'terraform')

    with open(shim_path, 'w') as open_file:
        open_file.write(shim_template.format(
            project_root=project_root,
            terraform_bin=terraform_bin,
        ))

    os.chmod(shim_path, 0o755)

    # Print the path to use with Direnv's PATH_add function.
    print(os.path.dirname(shim_path))


if __name__ == '__main__':
    install()



import os
import sys

from jinjaform import aws, files, git, log, root, terraform
from jinjaform.config import args, cwd, env, project_root, terraform_bin


commands_bypassed = (
    'fmt',
    'help',
    '-help',
    'version',
    '-version',
)

commands_forbidden_inside_target = (
    'push',
)

commands_using_backend = (
    'apply',
    'console',
    'debug'
    'destroy',
    'force-unlock',
    'graph',
    'import',
    'init',
    'output',
    'plan',
    'refresh',
    'show',
    'state',
    'taint',
    'untaint',
)

clean_after = False

try:

    if args:

        if not set(commands_bypassed).intersection(args):

            in_project_root = (cwd + '/').startswith(project_root + '/')
            in_target_path = in_project_root and os.path.exists('terraform.tfvars')

            if in_target_path:

                for command in commands_forbidden_inside_target:
                    if command in args:
                        log.bad('{} not allowed', command)
                        sys.exit(1)

                clean_after = True

                git.check()
                root.setup()
                files.delete()
                files.create()

                if set(commands_using_backend).intersection(args):
                    aws.credentials_setup()
                    if 'init' in args:
                        aws.backend_setup()

            else:

                log.bad('not in deployment target directory, aborting')
                sys.exit(1)

    log.ok('running terraform')
    sys.exit(terraform.execute(terraform_bin, args, env))

finally:

    if clean_after:
        files.delete()



import jinja2
import os
import shutil

from jinjaform import log, config
from jinjaform.config import cwd, jinjaform_dir, tf_vars


def create():
    """
    Creates symlinks and render templates from parent directories
    into the current directory.

    """

    # Discover files to create.
    seen = set()
    links = []
    templates = []
    current = cwd
    targets_path = os.path.realpath(os.path.join(os.environ['TERRAFORM_ROOT'], 'targets'))
    while (current + '/').startswith(targets_path + '/'):
        for name in sorted(os.listdir(current)):
            if name in seen:
                continue
            source = os.path.join(current, name)
            if os.path.isfile(source):
                if name.lower().endswith('.j2'):
                    name = name[:-3]
                    if name in seen:
                        continue
                    if not os.path.exists(name) and name:
                        templates.append((name, source))
                        seen.add(name)
                elif not os.path.exists(name):
                    links.append((name, source))
                    seen.add(name)

        current = os.path.dirname(current)

    # Create the symlinks.
    links.sort()
    for name, source in links:
        log.ok('link: {}', name)
        os.symlink(os.path.relpath(source), name)

    # Read the config files to find variables for rendering.
    config.read()

    # Render the templates.
    templates.sort()
    os.makedirs(jinjaform_dir, exist_ok=True)
    for name, source in templates:

        log.ok('render: {}', name)

        with open(source) as open_file:
            template = jinja2.Template(
                open_file.read(),
                undefined=jinja2.StrictUndefined,
            )

        context = os.environ.copy()
        context['var'] = tf_vars

        rendered = template.render(**context)

        output = os.path.join(jinjaform_dir, name)

        with open(output, 'w') as open_file:
            open_file.write(rendered)

        os.symlink(os.path.relpath(output), name)

    # Read again in case variables were inside templates.
    config.read()


def delete():
    """
    Deletes symlinks and rendered templates from the current directory.

    """

    for name in sorted(os.listdir()):
        if os.path.islink(name):
            os.remove(name)

    if os.path.exists(jinjaform_dir):
        shutil.rmtree(jinjaform_dir)



import errno
import os
import subprocess
import sys


def execute(terraform_bin, args, env):
    child_pid = os.fork()
    if child_pid == 0:
        os.execvpe(terraform_bin, [terraform_bin] + args, env)
    else:
        while True:
            try:
                _, exit_status = os.waitpid(child_pid, 0)
            except KeyboardInterrupt:
                pass
            except OSError as error:
                if error.errno == errno.ECHILD:
                    # No child processes.
                    # It has exited already.
                    break
                elif error.errno == errno.EINTR:
                    # Interrupted system call.
                    # This happens when resizing the terminal.
                    pass
                else:
                    # An actual error occurred.
                    raise
            else:
                return exit_status


def fmt(terraform_bin, name, _cache={}):
    """
    Runs `terraform fmt` on some input text and returns the result.

    """

    if name not in _cache:
        try:
            with open(name) as open_file:
                proc = subprocess.run(
                    [terraform_bin, 'fmt', '-'],
                    stdout=subprocess.PIPE,
                    input=open_file.read().encode('utf-8'),
                    check=True,
                )
        except subprocess.CalledProcessError as error:
            sys.exit(error.returncode)
        else:
            _cache[name] = proc.stdout.decode('utf-8')
    return _cache[name]
